import React, { createContext, useContext, useState, useRef, useEffect } from 'react';
import { select } from 'd3-selection';
import 'd3-transition';
import { easeExpOut } from 'd3-ease';
import { isNil, mapValues, pick, partition, clamp, range, omit, groupBy, keyBy, compact, get } from 'lodash-es';
import { curveBumpY, curveBumpX, curveStepAfter, curveStepBefore, curveStep, curveMonotoneY, curveMonotoneX, curveCatmullRom, curveLinear, area, line as line$1 } from 'd3-shape';
import { scalePoint, scaleBand, scaleTime, scalePow, scaleLog, scaleSqrt, scaleLinear, scaleOrdinal } from 'd3-scale';
import { line } from 'd3';
import { createPortal } from 'react-dom';

var DEFAULT_EASE = easeExpOut;

/* eslint-disable no-nested-ternary, no-underscore-dangle */

function mapKeys(obj, fn) {
  var entries = Object.entries(obj);
  var mapped = entries.map(function (_ref) {
    var k = _ref[0],
        v = _ref[1];
    return [fn(k), v];
  });
  return Object.fromEntries(mapped);
}

function parseAttributeName(str) {
  return str.replace(/([a-z0-9])([A-Z])/g, '$1-$2').replace(/([A-Z])([A-Z])(?=[a-z])/g, '$1-$2').toLowerCase();
}

function parseEventName(str) {
  return str.replace(/on(.*)/, '$1').toLowerCase();
}

function fallback() {
  for (var _len = arguments.length, fns = new Array(_len), _key = 0; _key < _len; _key++) {
    fns[_key] = arguments[_key];
  }

  return function () {
    for (var i = 0; i < fns.length; i++) {
      var fn = fns[i];
      var value = typeof fn === 'function' ? fn.apply(void 0, arguments) : fn;
      var isValidValue = value !== undefined && value !== null && (typeof value !== 'number' || !isNaN(value));
      if (isValidValue) return value;
    }

    return undefined;
  };
}

function AnimatedDataset(_ref2) {
  var dataset = _ref2.dataset,
      unparsedAttrs = _ref2.attrs,
      _ref2$tag = _ref2.tag,
      tag = _ref2$tag === void 0 ? 'rect' : _ref2$tag,
      _ref2$init = _ref2.init,
      unparsedInit = _ref2$init === void 0 ? {} : _ref2$init,
      _ref2$events = _ref2.events,
      unparsedEvents = _ref2$events === void 0 ? {} : _ref2$events,
      _ref2$keyFn = _ref2.keyFn,
      keyFn = _ref2$keyFn === void 0 ? function (d) {
    return d;
  } : _ref2$keyFn,
      _ref2$duration = _ref2.duration,
      duration = _ref2$duration === void 0 ? 500 : _ref2$duration,
      _ref2$delay = _ref2.delay,
      delay = _ref2$delay === void 0 ? 0 : _ref2$delay,
      _ref2$tweens = _ref2.tweens,
      unparsedTweens = _ref2$tweens === void 0 ? {} : _ref2$tweens,
      _ref2$disableAnimatio = _ref2.disableAnimation,
      disableAnimation = _ref2$disableAnimatio === void 0 ? false : _ref2$disableAnimatio,
      _ref2$easing = _ref2.easing,
      easing = _ref2$easing === void 0 ? DEFAULT_EASE : _ref2$easing;
  var ref = React.createRef();
  var refOldAttrs = React.useRef();
  React.useLayoutEffect(function () {
    if (!ref.current) return;
    var attrs = mapKeys(unparsedAttrs, parseAttributeName);
    var init = mapKeys(unparsedInit, parseAttributeName);
    var events = mapKeys(unparsedEvents, parseEventName);
    var tweens = mapKeys(unparsedTweens, parseAttributeName);
    var attrsList = Object.keys(attrs).filter(function (a) {
      return a !== 'text';
    });
    var tweensList = Object.keys(tweens).filter(function (a) {
      return a !== 'text';
    });
    var eventsList = Object.keys(events);
    var oldAttrs = refOldAttrs.current || {};
    var oldAttrsList = Object.keys(oldAttrs).filter(function (a) {
      return a !== 'text';
    });

    var animate = function animate() {
      select(ref.current).selectAll(tag).data(dataset, keyFn).join(function (enter) {
        return enter.append(tag).text(attrs.text).call(function (sel) {
          attrsList.forEach(function (a) {
            sel.attr(a, fallback(init[a], oldAttrs[a], attrs[a]));
          });
        }).call(function (sel) {
          eventsList.forEach(function (event) {
            sel.on(event, events[event]);
          });
        }).call(function (sel) {
          var tran = disableAnimation ? sel : sel.transition().ease(easing).delay(delay).duration(duration);
          attrsList.forEach(function (a) {
            tran.attr(a, attrs[a]);
          });
          tweensList.forEach(function (a) {
            tran.attrTween(a, tweens[a]);
          });
          tran.textTween(tweens.text);
        });
      }, function (update) {
        return update.text(attrs.text).call(function (sel) {
          eventsList.forEach(function (event) {
            sel.on(event, events[event]);
          });
        }).call(function (sel) {
          var tran = disableAnimation ? sel : sel.transition().ease(easing).delay(delay).duration(duration);
          attrsList.forEach(function (a) {
            tran.attr(a, attrs[a]);
          });
          tweensList.forEach(function (a) {
            tran.attrTween(a, tweens[a]);
          });
          tran.textTween(tweens.text);
        });
      }, function (exit) {
        return exit.call(function (sel) {
          var tran = disableAnimation ? sel : sel.transition().ease(easing).delay(delay).duration(duration);
          oldAttrsList.forEach(function (a) {
            tran.attr(a, fallback(init[a], oldAttrs[a]));
          });
          tweensList.forEach(function (a) {
            tran.attrTween(a, tweens[a]);
          });
          tran.textTween(tweens.text);
          tran.remove();
        });
      });
      refOldAttrs.current = attrs;
    };

    if (disableAnimation) {
      animate();
    } else {
      requestAnimationFrame(animate);
    }
  }, [unparsedTweens, easing, dataset, unparsedInit, keyFn, ref, tag, unparsedAttrs, duration, disableAnimation, unparsedEvents, delay]);
  return React.createElement('g', {
    ref: ref
  });
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

var tuple = function tuple() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return args;
};
var strTuple = function strTuple() {
  for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    args[_key2] = arguments[_key2];
  }

  return args;
};
var isUndefinedTuple = function isUndefinedTuple(x) {
  return x.length === 2 && x[0] === undefined && x[1] === undefined;
};
var isNotNil = function isNotNil(t) {
  return t !== null && t !== undefined;
};
function wrapIteratee(iteratee, map) {
  return function (d, i) {
    return iteratee(map(d), i);
  };
}
function isAccessor(getter) {
  return typeof getter === 'function';
}
function toAccessor(getter) {
  return isAccessor(getter) ? getter : function () {
    return getter;
  };
}
var isScaleContinuous = function isScaleContinuous(v) {
  return typeof v === 'function' && typeof v.invert === 'function' && typeof v.nice === 'function' && typeof v.copy === 'function';
};
var isScaleCategorical = function isScaleCategorical(v) {
  return typeof v === 'function' && typeof v.copy === 'function' && typeof v.bandwidth === 'function';
};

var xand = function xand(a, b) {
  return a === b;
};
var xor = function xor(a, b) {
  return !xand(a, b);
};
var removeEmptyKeys = function removeEmptyKeys(o) {
  return Object.fromEntries(Object.entries(o).filter(function (_ref) {
    var v = _ref[1];
    return !isNil(v);
  }));
};

function buildContext(dependency) {
  var CurrentContext = createContext(null);

  var useHook = function useHook() {
    var value = useContext(CurrentContext);
    if (value === null && dependency !== undefined) throw new Error(dependency + " missing");
    return value;
  };

  var Consumer = function Consumer(_ref) {
    var children = _ref.children;
    var value = useHook();
    return React.createElement(React.Fragment, null, children(value));
  };

  return tuple(CurrentContext, useHook, Consumer);
}

var _buildContext = /*#__PURE__*/buildContext('Chart'),
    ChartContext = _buildContext[0],
    useChartContext = _buildContext[1],
    ChartConsumer = _buildContext[2];

var _buildContext2 = /*#__PURE__*/buildContext('Grid'),
    GridContext = _buildContext2[0],
    useGridContext = _buildContext2[1],
    GridConsumer = _buildContext2[2];

var _buildContext3 = /*#__PURE__*/buildContext(),
    StyleContext = _buildContext3[0],
    useStyleContext = _buildContext3[1],
    StyleConsumer = _buildContext3[2];

var _buildContext4 = /*#__PURE__*/buildContext(),
    AnimationContext = _buildContext4[0],
    useAnimationContext = _buildContext4[1],
    AnimationConsumer = _buildContext4[2];

var _buildContext5 = /*#__PURE__*/buildContext('Cartesian'),
    CartesianContext = _buildContext5[0],
    useCartesianContext = _buildContext5[1],
    CartesianConsumer = _buildContext5[2];

var _buildContext6 = /*#__PURE__*/buildContext("Interaction missing (use 'interactive' prop)"),
    MouseContext = _buildContext6[0],
    useMouseContext = _buildContext6[1],
    MouseConsumer = _buildContext6[2];

var _excluded = ["easing"],
    _excluded2 = ["children"];

function sanitizeAnimationOptions(_ref) {
  var easing = _ref.easing,
      obj = _objectWithoutPropertiesLoose(_ref, _excluded);

  var sanitized = mapValues(obj, function (value) {
    return typeof value === 'function' ? wrapIteratee(value, function (d) {
      return d.datum;
    }) : value;
  });
  return _extends({}, sanitized, {
    easing: easing
  });
}

function useCascadingAnimation(curr) {
  var _useAnimationContext;

  if (curr === void 0) {
    curr = {};
  }

  var animation = _extends({}, (_useAnimationContext = useAnimationContext()) != null ? _useAnimationContext : {
    delay: 0,
    duration: 500,
    easing: DEFAULT_EASE
  }, removeEmptyKeys(curr));

  return animation;
}
function useSanitizedCascadingAnimation(anim) {
  return sanitizeAnimationOptions(useCascadingAnimation(anim));
}
function Animation(_ref2) {
  var children = _ref2.children,
      animationProps = _objectWithoutPropertiesLoose(_ref2, _excluded2);

  var animation = useCascadingAnimation(animationProps);
  return React.createElement(AnimationContext.Provider, {
    value: animation
  }, children);
}

function buildCurveFn(type) {
  switch (type) {
    case 'line':
      return curveLinear;

    case 'curve':
      return curveCatmullRom.alpha(1);

    case 'monotone-x':
      return curveMonotoneX;

    case 'monotone-y':
      return curveMonotoneY;

    case 'step':
      return curveStep;

    case 'step-before':
      return curveStepBefore;

    case 'step-after':
      return curveStepAfter;

    case 'bump-x':
      return curveBumpX;

    case 'bump-y':
      return curveBumpY;

    default:
      throw new Error("Unknown curve type " + type);
  }
}

var scaleZero = function scaleZero(scale) {
  return isScaleCategorical(scale) ? scale.domain()[0] : 0;
};
var scaleStep = function scaleStep(scale) {
  return isScaleCategorical(scale) ? scale.bandwidth() : 1;
};

function computeCenterPos(value, scale) {
  var pos = scale(value);
  var off = isScaleCategorical(scale) ? scale.bandwidth() / 2 : 0;
  return pos + off;
}

function computeStartPos(value, scale) {
  var pos = scale(value);
  return pos;
}

function computeEndPos(value, scale) {
  var pos = scale(value);
  return pos + scaleStep(scale);
}

function computePos(value, scale, position) {
  if (position === void 0) {
    position = 'center';
  }

  if (position === 'center') return computeCenterPos(value, scale);
  if (position === 'start') return computeStartPos(value, scale);
  if (position === 'end') return computeEndPos(value, scale);
  return computeStartPos(value, scale) + scaleStep(scale) * position;
}
function createDataScale(dataAccessor, scale, defaultValue, position) {
  if (position === void 0) {
    position = 'center';
  }

  return function (d) {
    var computed = dataAccessor ? computePos(dataAccessor(d), scale, position) : defaultValue;
    return computed;
  };
}
function combineGetters(getters) {
  return function (d) {
    return getters.filter(isNotNil).map(function (g) {
      return toAccessor(g)(d);
    }).reduce(function (a, b) {
      return a + b;
    }, 0);
  };
}

var _excluded$1 = ["children"];
function useCascadingStyle(curr) {
  var _useStyleContext;

  return _extends({}, (_useStyleContext = useStyleContext()) != null ? _useStyleContext : {}, curr);
}
function Style(_ref) {
  var children = _ref.children,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$1);

  var style = useCascadingStyle(props);
  return React.createElement(StyleContext.Provider, {
    value: style
  }, children);
}

var _excluded$2 = ["tag", "data", "dataKey", "delay", "duration", "enter", "easing"];

var splitSvgProps = function splitSvgProps(props) {
  var keys = Object.keys(props);
  var existingKeys = keys.filter(function (k) {
    return typeof k !== 'undefined';
  });

  var _partition = partition(existingKeys, function (key) {
    return key.match(/^on[A-Z]/) !== null;
  }),
      fnKeys = _partition[0],
      valKeys = _partition[1];

  return tuple(pick(props, valKeys), pick(props, fnKeys));
};

function Elements(_ref) {
  var tag = _ref.tag,
      data = _ref.data,
      dataKey = _ref.dataKey,
      delay = _ref.delay,
      duration = _ref.duration,
      enter = _ref.enter,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$2);

  var animation = useSanitizedCascadingAnimation({
    delay: delay,
    duration: duration,
    easing: easing
  });

  var _splitSvgProps = splitSvgProps(props),
      svgProps = _splitSvgProps[0],
      events = _splitSvgProps[1];

  var attributes = _extends({}, useCascadingStyle({}), svgProps);

  return React.createElement(AnimatedDataset, Object.assign({
    tag: tag,
    dataset: data,
    keyFn: function keyFn(d, i) {
      var _dataKey;

      return (_dataKey = dataKey == null ? void 0 : dataKey(d, i)) != null ? _dataKey : i;
    },
    attrs: attributes,
    events: events,
    init: enter
  }, animation));
}

var _excluded$3 = ["data", "x-data", "y-data", "curve", "x", "y1", "y2"];

function sanitizeAxesConfig(config, scale) {
  if (typeof config === 'function') {
    return {
      to: config,
      base: function base() {
        return scaleZero(scale);
      }
    };
  }

  var base = config.base,
      to = config.to;
  return {
    to: to,
    base: toAccessor(base)
  };
}

function Area(_ref) {
  var data = _ref.data,
      dataX = _ref['x-data'],
      _dataY = _ref['y-data'],
      _ref$curve = _ref.curve,
      curve = _ref$curve === void 0 ? 'line' : _ref$curve,
      _x = _ref.x,
      _y1 = _ref.y1,
      _y2 = _ref.y2,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$3);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var curveFn = buildCurveFn(curve);
  var dataY = sanitizeAxesConfig(_dataY, yScale);
  var x = combineGetters([createDataScale(dataX, xScale, 0), _x]);
  var y1 = combineGetters([createDataScale(dataY.base, yScale, 0), _y1]);
  var y2 = combineGetters([createDataScale(dataY.to, yScale, 0), _y2]);

  var d = function d(data) {
    var _area$x$y0$y1$curve;

    var points = data.map(function (d) {
      return [x(d), y1(d), y2(d)];
    }).filter(function (d) {
      return d.every(function (p) {
        return !isNaN(p);
      });
    });
    return (_area$x$y0$y1$curve = area().x(function (d) {
      return d[0];
    }).y0(function (d) {
      return d[1];
    }).y1(function (d) {
      return d[2];
    }).curve(curveFn)(points)) != null ? _area$x$y0$y1$curve : '';
  };

  return React.createElement(Elements, Object.assign({
    tag: "path",
    data: [data]
  }, props, {
    d: d
  }));
}

var _excluded$4 = ["x-data", "y-data"];

function sanitizeAxisConfig(config, scale) {
  if (typeof config === 'function') {
    return {
      to: config,
      base: config
    };
  }

  var to = config.to,
      _config$base = config.base,
      base = _config$base === void 0 ? function () {
    return scaleZero(scale);
  } : _config$base;
  return {
    to: toAccessor(to),
    base: toAccessor(base)
  };
}

function Bars(_ref) {
  var _dataX = _ref['x-data'],
      _dataY = _ref['y-data'],
      props = _objectWithoutPropertiesLoose(_ref, _excluded$4);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var dataX = sanitizeAxisConfig(_dataX, xScale);
  var dataY = sanitizeAxisConfig(_dataY, yScale);
  var x1 = createDataScale(dataX.base, xScale, 0, 'start');
  var y1 = createDataScale(dataY.base, yScale, 0, 'end');
  var x2 = createDataScale(dataX.to, xScale, 0, 'end');
  var y2 = createDataScale(dataY.to, yScale, 0, 'start');

  var x = function x(d) {
    return Math.min(x1(d), x2(d));
  };

  var y = function y(d) {
    return Math.min(y1(d), y2(d));
  };

  var width = function width(d) {
    return Math.abs(x1(d) - x2(d));
  };

  var height = function height(d) {
    return Math.abs(y1(d) - y2(d));
  };

  return React.createElement(Elements, Object.assign({}, props, {
    tag: "rect",
    x: combineGetters([x, props.x]),
    y: combineGetters([y, props.y]),
    width: combineGetters([width, props.width]),
    height: combineGetters([height, props.height])
  }));
}

/* eslint-disable indent */

function buildScale(options, range) {
  var _options$domain, _options$domain2, _options$domain3, _options$base, _options$domain4, _options$exponent, _options$domain5, _options$domain7, _options$padding;

  switch (options.scale) {
    case 'linear':
      return scaleLinear().domain((_options$domain = options == null ? void 0 : options.domain) != null ? _options$domain : [0, 1]).range(range);

    case 'square':
      return scaleSqrt().domain((_options$domain2 = options == null ? void 0 : options.domain) != null ? _options$domain2 : [0, 1]).range(range);

    case 'log':
      return scaleLog().domain((_options$domain3 = options == null ? void 0 : options.domain) != null ? _options$domain3 : [0, 1]).range(range).base((_options$base = options.base) != null ? _options$base : 10);

    case 'pow':
      return scalePow().domain((_options$domain4 = options == null ? void 0 : options.domain) != null ? _options$domain4 : [0, 1]).range(range).exponent((_options$exponent = options.exponent) != null ? _options$exponent : 1);

    case 'date':
      return scaleTime().domain((_options$domain5 = options == null ? void 0 : options.domain) != null ? _options$domain5 : []).range(range);

    case 'band':
      {
        var _options$domain6;

        // TODO: add padding, paddingInner, paddingOuter options based on pixels
        var scale = scaleBand((_options$domain6 = options == null ? void 0 : options.domain) != null ? _options$domain6 : [], range);
        if (options.padding !== undefined) scale.padding(options.padding);
        if (options.paddingInner !== undefined) scale.paddingInner(options.paddingInner);
        if (options.paddingOuter !== undefined) scale.paddingOuter(options.paddingOuter);
        return scale;
      }

    case 'point':
      // TODO: add padding option based on pixels
      return scalePoint().domain((_options$domain7 = options == null ? void 0 : options.domain) != null ? _options$domain7 : []).range(range).padding((_options$padding = options == null ? void 0 : options.padding) != null ? _options$padding : 0);

    default:
      throw new Error("Unknown scale options: " + options);
  }
}

function buildOrdinalScale(options) {
  switch (options.scale) {
    case 'ordinal':
      return scaleOrdinal().domain(options.domain).range(options.range);

    default:
      throw new Error("Unknown scale options: " + options);
  }
}

function Cartesian(_ref) {
  var x = _ref.x,
      y = _ref.y,
      color = _ref.color,
      children = _ref.children,
      _ref$nice = _ref.nice,
      nice = _ref$nice === void 0 ? false : _ref$nice,
      _ref$interactive = _ref.interactive,
      interactive = _ref$interactive === void 0 ? false : _ref$interactive;

  var _useChartContext = useChartContext(),
      top = _useChartContext.top,
      left = _useChartContext.left,
      bottom = _useChartContext.bottom,
      right = _useChartContext.right;

  var xRange = tuple(left, right);
  var yRange = tuple(bottom, top);
  var xScaleNotNice = buildScale(x != null ? x : {
    scale: 'linear'
  }, xRange);
  var yScaleNotNice = buildScale(y != null ? y : {
    scale: 'linear'
  }, yRange);
  var colorScale = color ? buildOrdinalScale(color) : undefined;
  var xScale = xScaleNotNice.copy();
  var yScale = yScaleNotNice.copy();
  if ((nice === true || nice === 'x') && isScaleContinuous(xScale)) xScale.nice();
  if ((nice === true || nice === 'y') && isScaleContinuous(yScale)) yScale.nice();
  return React.createElement(CartesianContext.Provider, {
    value: {
      xScale: xScale,
      yScale: yScale,
      colorScale: colorScale
    }
  }, React.createElement(MouseListener, {
    interactive: interactive
  }, children));
}

var MouseListener = function MouseListener(_ref2) {
  var interactive = _ref2.interactive,
      children = _ref2.children;

  var _useChartContext2 = useChartContext(),
      top = _useChartContext2.top,
      left = _useChartContext2.left,
      width = _useChartContext2.width,
      height = _useChartContext2.height;

  var _useState = useState(undefined),
      mouse = _useState[0],
      setMouse = _useState[1];

  var handlePointerMove = function handlePointerMove(e) {
    var bbox = e.currentTarget.getBoundingClientRect();
    var x = e.clientX - bbox.left + left;
    var y = e.clientY - bbox.top + top;
    setMouse({
      x: x,
      y: y
    });
  };

  var handlePointerLeave = function handlePointerLeave() {
    setMouse(undefined);
  };

  return React.createElement(MouseContext.Provider, {
    value: mouse
  }, children, interactive && React.createElement("rect", {
    x: left,
    y: top,
    width: width,
    height: height,
    fill: "transparent",
    style: {
      touchAction: 'none'
    },
    onPointerMove: handlePointerMove,
    onPointerLeave: handlePointerLeave
  }));
};

function Chart(_ref) {
  var width = _ref.width,
      height = _ref.height,
      children = _ref.children,
      _ref$top = _ref.top,
      top = _ref$top === void 0 ? 0 : _ref$top,
      _ref$left = _ref.left,
      left = _ref$left === void 0 ? 0 : _ref$left;
  return React.createElement(ChartContext.Provider, {
    value: {
      width: width,
      height: height,
      top: top,
      left: left,
      right: width + left,
      bottom: height + top
    }
  }, children);
}

var _excluded$5 = ["x-data", "y-data"];
function Circles(_ref) {
  var dataX = _ref['x-data'],
      dataY = _ref['y-data'],
      props = _objectWithoutPropertiesLoose(_ref, _excluded$5);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var cx = createDataScale(dataX, xScale, 0);
  var cy = createDataScale(dataY, yScale, 0);
  return React.createElement(Elements, Object.assign({}, props, {
    tag: "circle",
    cx: combineGetters([cx, props.cx]),
    cy: combineGetters([cy, props.cy])
  }));
}

var useConst = function useConst(initial) {
  return useRef(initial).current;
};

var useId = function useId() {
  return useConst(Math.random().toString(16).split('.')[1]);
};

function ClipRect(_ref) {
  var width = _ref.width,
      height = _ref.height,
      _ref$top = _ref.top,
      top = _ref$top === void 0 ? 0 : _ref$top,
      _ref$left = _ref.left,
      left = _ref$left === void 0 ? 0 : _ref$left,
      children = _ref.children;
  var id = useId();
  return React.createElement(React.Fragment, null, React.createElement("defs", null, React.createElement("clipPath", {
    id: id
  }, React.createElement("rect", {
    x: left,
    y: top,
    width: width,
    height: height
  }))), React.createElement("g", {
    clipPath: "url(#" + id + ")"
  }, children));
}

function ClipChart(_ref) {
  var children = _ref.children;
  return React.createElement(ClipRect, Object.assign({}, useChartContext()), children);
}

var _excluded$6 = ["duration", "delay", "easing", "tweens"];
var tags = /*#__PURE__*/strTuple('rect', 'line', 'circle', 'path', 'text', 'polygon', 'tspan', 'ellipse', 'g');
var element = /*#__PURE__*/Object.fromEntries( /*#__PURE__*/tags.map(function (Tag) {
  return [Tag, // TODO: add forward ref
  function (_ref) {
    var duration = _ref.duration,
        delay = _ref.delay,
        easing = _ref.easing,
        tweens = _ref.tweens,
        props = _objectWithoutPropertiesLoose(_ref, _excluded$6);

    return React.createElement(AnimatedDataset, Object.assign({
      tag: Tag,
      dataset: [{}],
      keyFn: function keyFn(_, i) {
        return i;
      },
      attrs: useCascadingStyle(props),
      tweens: tweens
    }, useCascadingAnimation({
      duration: duration,
      delay: delay,
      easing: easing
    })));
  }];
}));

var clampValue = function clampValue(value, _ref) {
  var min = _ref[0],
      max = _ref[1];

  if (min > max) {
    var _ref2 = [max, min];
    min = _ref2[0];
    max = _ref2[1];
  }

  return clamp(value, min, max);
};

var applyTransform = function applyTransform(range, transform) {
  var length = Math.abs(range[0] - range[1]);
  var n0 = transform[0] * length + range[0];
  var n1 = transform[1] * length + range[1];
  return [n0, n1];
};

var inferTransform = function inferTransform(prev, curr) {
  var length = Math.abs(prev[0] - prev[1]);
  var t0 = (curr[0] - prev[0]) / length;
  var t1 = (curr[1] - prev[1]) / length;
  return [t0, t1];
};

var Range = {
  clampValue: clampValue,
  inferTransform: inferTransform,
  applyTransform: applyTransform
};

var _excluded$7 = ["format", "padding", "threshold", "inner", "filter", "ticks", "dataKey", "enter", "delay", "duration"],
    _excluded2$1 = ["format", "padding", "threshold", "inner", "filter", "ticks", "dataKey", "delay", "duration", "enter"],
    _excluded3 = ["ticks", "dataKey", "delay", "duration", "enter", "filter"],
    _excluded4 = ["ticks", "dataKey", "enter", "duration", "delay", "filter"];

var computeAxesPos = function computeAxesPos(scale, range, anchor) {
  if (anchor === void 0) {
    anchor = 'none';
  }

  var _start = range[0],
      _end = range[1];
  return {
    start: function start() {
      return _start;
    },
    end: function end() {
      return _end;
    },
    none: function none() {
      var zeroPos = isScaleContinuous(scale) ? scale(0) : scale.range()[0];
      return Range.clampValue(zeroPos, [_start, _end]);
    }
  }[anchor]();
};

var computeTicks = function computeTicks(scale, count) {
  return isScaleContinuous(scale) ? scale.ticks(count) : scale.domain();
};

var Grid = function Grid(_ref) {
  var children = _ref.children,
      _ref$tickSize = _ref.tickSize,
      tickResolution = _ref$tickSize === void 0 ? 50 : _ref$tickSize,
      tickCount = _ref.tickCount,
      _ref$xAnchor = _ref.xAnchor,
      xAnchor = _ref$xAnchor === void 0 ? 'none' : _ref$xAnchor,
      _ref$yAnchor = _ref.yAnchor,
      yAnchor = _ref$yAnchor === void 0 ? 'none' : _ref$yAnchor;

  var _useChartContext = useChartContext(),
      top = _useChartContext.top,
      left = _useChartContext.left,
      bottom = _useChartContext.bottom,
      right = _useChartContext.right,
      width = _useChartContext.width,
      height = _useChartContext.height;

  var _useCartesianContext = useCartesianContext(),
      yScale = _useCartesianContext.yScale,
      xScale = _useCartesianContext.xScale;

  var xCount = tickCount != null ? tickCount : width / tickResolution;
  var yCount = tickCount != null ? tickCount : height / tickResolution;
  var xAxesAnchor = {
    bottom: 'start',
    top: 'end',
    none: 'none'
  }[xAnchor];
  var yAxesAnchor = {
    left: 'start',
    right: 'end',
    none: 'none'
  }[yAnchor];
  return React.createElement(GridContext.Provider, {
    value: {
      xAxes: computeAxesPos(xScale, [left, right], yAxesAnchor),
      yAxes: computeAxesPos(yScale, [bottom, top], xAxesAnchor),
      xTicks: computeTicks(xScale, xCount),
      yTicks: computeTicks(yScale, yCount)
    }
  }, children);
};

Grid.XAxes = function (props) {
  var _useChartContext2 = useChartContext(),
      left = _useChartContext2.left,
      right = _useChartContext2.right;

  var _useGridContext = useGridContext(),
      yAxes = _useGridContext.yAxes; // TODO: add animation


  return React.createElement(element.line, Object.assign({}, props, {
    x1: left,
    x2: right,
    y1: yAxes,
    y2: yAxes
  }));
};

Grid.YAxes = function (props) {
  var _useChartContext3 = useChartContext(),
      top = _useChartContext3.top,
      bottom = _useChartContext3.bottom;

  var _useGridContext2 = useGridContext(),
      xAxes = _useGridContext2.xAxes; // TODO: add animation


  return React.createElement(element.line, Object.assign({}, props, {
    x1: xAxes,
    x2: xAxes,
    y1: top,
    y2: bottom
  }));
};

Grid.XLabels = function (_ref2) {
  var _ref2$format = _ref2.format,
      buildFormatter = _ref2$format === void 0 ? function () {
    return function (value) {
      return String(value);
    };
  } : _ref2$format,
      _ref2$padding = _ref2.padding,
      padding = _ref2$padding === void 0 ? 0 : _ref2$padding,
      _ref2$threshold = _ref2.threshold,
      threshold = _ref2$threshold === void 0 ? 0 : _ref2$threshold,
      _ref2$inner = _ref2.inner,
      _inner = _ref2$inner === void 0 ? false : _ref2$inner,
      _ref2$filter = _ref2.filter,
      filter = _ref2$filter === void 0 ? function () {
    return true;
  } : _ref2$filter,
      _ticks = _ref2.ticks,
      _dataKey = _ref2.dataKey,
      _ref2$enter = _ref2.enter,
      enter = _ref2$enter === void 0 ? {} : _ref2$enter,
      props = _objectWithoutPropertiesLoose(_ref2, _excluded$7);

  var _useChartContext4 = useChartContext(),
      top = _useChartContext4.top;

  var _useGridContext3 = useGridContext(),
      yAxes = _useGridContext3.yAxes,
      xTicks = _useGridContext3.xTicks;

  var ticks = _ticks != null ? _ticks : xTicks;
  var filteredTicks = ticks.filter(filter);
  var reverse = yAxes <= top + threshold;
  var inner = xor(_inner, reverse);
  var formatter = buildFormatter(filteredTicks);
  return React.createElement(Texts, Object.assign({
    data: filteredTicks,
    "x-data": function xData(t) {
      return t;
    },
    y: yAxes + (inner ? -padding : padding),
    textAnchor: "middle",
    dominantBaseline: inner ? 'auto' : 'hanging',
    text: formatter,
    opacity: 1,
    enter: _extends({
      opacity: 0
    }, enter),
    dataKey: function dataKey(tick, i) {
      var _dataKey2;

      return (_dataKey2 = _dataKey == null ? void 0 : _dataKey(tick, i)) != null ? _dataKey2 : tick;
    }
  }, props));
};

Grid.YLabels = function (_ref3) {
  var _ref3$format = _ref3.format,
      buildFormatter = _ref3$format === void 0 ? function () {
    return function (value) {
      return String(value);
    };
  } : _ref3$format,
      _ref3$padding = _ref3.padding,
      padding = _ref3$padding === void 0 ? 0 : _ref3$padding,
      _ref3$threshold = _ref3.threshold,
      threshold = _ref3$threshold === void 0 ? 0 : _ref3$threshold,
      _ref3$inner = _ref3.inner,
      _inner = _ref3$inner === void 0 ? false : _ref3$inner,
      _ref3$filter = _ref3.filter,
      filter = _ref3$filter === void 0 ? function () {
    return true;
  } : _ref3$filter,
      _ticks = _ref3.ticks,
      _dataKey3 = _ref3.dataKey,
      enter = _ref3.enter,
      props = _objectWithoutPropertiesLoose(_ref3, _excluded2$1);

  var _useChartContext5 = useChartContext(),
      right = _useChartContext5.right;

  var _useGridContext4 = useGridContext(),
      xAxes = _useGridContext4.xAxes,
      yTicks = _useGridContext4.yTicks;

  var ticks = _ticks != null ? _ticks : yTicks;
  var filteredTicks = ticks.filter(filter);
  var reverse = xAxes >= right - threshold;
  var inner = xor(_inner, reverse);
  var formatter = buildFormatter(filteredTicks);
  return React.createElement(Texts, Object.assign({
    data: filteredTicks,
    x: xAxes + (inner ? padding : -padding),
    "y-data": function yData(t) {
      return t;
    },
    textAnchor: inner ? 'start' : 'end',
    dominantBaseline: "middle",
    text: formatter,
    opacity: 1,
    enter: _extends({
      opacity: 0
    }, enter),
    dataKey: function dataKey(tick, i) {
      var _dataKey4;

      return (_dataKey4 = _dataKey3 == null ? void 0 : _dataKey3(tick, i)) != null ? _dataKey4 : tick;
    }
  }, props));
};

Grid.XLines = function (_ref4) {
  var ticks = _ref4.ticks,
      _dataKey5 = _ref4.dataKey,
      enter = _ref4.enter,
      _ref4$filter = _ref4.filter,
      filter = _ref4$filter === void 0 ? function () {
    return true;
  } : _ref4$filter,
      props = _objectWithoutPropertiesLoose(_ref4, _excluded3);

  var _useChartContext6 = useChartContext(),
      top = _useChartContext6.top,
      bottom = _useChartContext6.bottom;

  var _useCartesianContext2 = useCartesianContext(),
      xScale = _useCartesianContext2.xScale;

  var xTicks = ticks != null ? ticks : useGridContext().xTicks;
  var filteredTicks = xTicks.filter(filter);
  return React.createElement(Elements, Object.assign({
    tag: "line",
    data: filteredTicks,
    x1: function x1(t) {
      return computePos(t, xScale);
    },
    x2: function x2(t) {
      return computePos(t, xScale);
    },
    y1: top,
    y2: bottom,
    opacity: 1,
    enter: _extends({
      opacity: 0
    }, enter),
    dataKey: function dataKey(tick, i) {
      var _dataKey6;

      return (_dataKey6 = _dataKey5 == null ? void 0 : _dataKey5(tick, i)) != null ? _dataKey6 : tick;
    }
  }, props));
};

Grid.YLines = function (_ref5) {
  var ticks = _ref5.ticks,
      _dataKey7 = _ref5.dataKey,
      enter = _ref5.enter,
      _ref5$filter = _ref5.filter,
      filter = _ref5$filter === void 0 ? function () {
    return true;
  } : _ref5$filter,
      props = _objectWithoutPropertiesLoose(_ref5, _excluded4);

  var _useChartContext7 = useChartContext(),
      left = _useChartContext7.left,
      right = _useChartContext7.right;

  var _useCartesianContext3 = useCartesianContext(),
      yScale = _useCartesianContext3.yScale;

  var yTicks = ticks != null ? ticks : useGridContext().yTicks;
  var filteredTicks = yTicks.filter(filter);
  return React.createElement(Elements, Object.assign({
    tag: "line",
    data: filteredTicks,
    x1: left,
    x2: right,
    y1: function y1(t) {
      return computePos(t, yScale);
    },
    y2: function y2(t) {
      return computePos(t, yScale);
    },
    opacity: 1,
    enter: _extends({
      opacity: 0
    }, enter),
    dataKey: function dataKey(tick, i) {
      var _dataKey8;

      return (_dataKey8 = _dataKey7 == null ? void 0 : _dataKey7(tick, i)) != null ? _dataKey8 : tick;
    }
  }, props));
};

var _excluded$8 = ["data", "x-data", "y-data", "x", "y", "curve"];
function Line(_ref) {
  var data = _ref.data,
      dataX = _ref['x-data'],
      dataY = _ref['y-data'],
      _x = _ref.x,
      _y = _ref.y,
      _ref$curve = _ref.curve,
      curveType = _ref$curve === void 0 ? 'line' : _ref$curve,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$8);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var curveFn = buildCurveFn(curveType);
  var x = combineGetters([createDataScale(dataX, xScale, 0), _x]);
  var y = combineGetters([createDataScale(dataY, yScale, 0), _y]);

  var d = function d(data) {
    var _line$curve;

    var points = data.map(function (d) {
      return tuple(x(d), y(d));
    }).filter(function (d) {
      return d.every(function (p) {
        return !isNaN(p);
      });
    });
    return (_line$curve = line().curve(curveFn)(points)) != null ? _line$curve : '';
  };

  return React.createElement(Elements, Object.assign({
    fill: "none"
  }, props, {
    data: [data],
    tag: "path",
    d: d
  }));
}

function Portal(_ref) {
  var children = _ref.children,
      container = _ref.container;
  return createPortal(children, container);
}

var _excluded$9 = ["x-data", "y-data", "data-width", "data-height"];
function Rects(_ref) {
  var dataX = _ref['x-data'],
      dataY = _ref['y-data'],
      dataWidth = _ref['data-width'],
      dataHeight = _ref['data-height'],
      props = _objectWithoutPropertiesLoose(_ref, _excluded$9);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var x = createDataScale(dataX, xScale, 0);
  var y = createDataScale(dataY, yScale, 0);
  var width = createDataScale(dataWidth, xScale, 0);
  var height = createDataScale(dataHeight, yScale, 0);
  return React.createElement(Elements, Object.assign({}, props, {
    tag: "rect",
    x: combineGetters([x, props.x]),
    y: combineGetters([y, props.y]),
    width: combineGetters([width, props.width]),
    height: combineGetters([height, props.height])
  }));
}

function useForceRender() {
  var _useState = useState(false),
      rendered = _useState[0],
      setRendered = _useState[1];

  useEffect(function () {
    if (!rendered) setRendered(true);
  }, [rendered]);
}

var TooltipParentContext = /*#__PURE__*/createContext(null);
var TooltipParentProvider = TooltipParentContext.Provider;
function useTooltipParent() {
  var tooltipParent = useContext(TooltipParentContext);
  if (!tooltipParent) throw Error('use Svg component');
  return tooltipParent;
}
var ANCHOR_OFFSET = {
  start: '',
  center: '- 50%',
  end: '- 100%'
};
function Tooltip(_ref) {
  var x = _ref.x,
      y = _ref.y,
      _ref$children = _ref.children,
      children = _ref$children === void 0 ? null : _ref$children,
      _ref$style = _ref.style,
      style = _ref$style === void 0 ? {} : _ref$style,
      _ref$horizontalAnchor = _ref.horizontalAnchor,
      horizontalAnchor = _ref$horizontalAnchor === void 0 ? 'start' : _ref$horizontalAnchor,
      _ref$verticalAnchor = _ref.verticalAnchor,
      verticalAnchor = _ref$verticalAnchor === void 0 ? 'start' : _ref$verticalAnchor,
      _ref$horizontalSpacin = _ref.horizontalSpacing,
      horizontalSpacing = _ref$horizontalSpacin === void 0 ? 0 : _ref$horizontalSpacin,
      _ref$verticalSpacing = _ref.verticalSpacing,
      verticalSpacing = _ref$verticalSpacing === void 0 ? 0 : _ref$verticalSpacing;
  var parent = useTooltipParent();
  return React.createElement(Portal, {
    container: parent
  }, React.createElement("div", {
    style: {
      position: 'absolute',
      left: 0,
      top: 0,
      transform: "translate(" + x + "px, " + y + "px)",
      pointerEvents: 'none',
      zIndex: 10
    }
  }, React.createElement("div", {
    style: _extends({
      transition: 'transform 0.15s',
      transform: "translate(\n            calc(" + horizontalSpacing + "px " + ANCHOR_OFFSET[horizontalAnchor] + "),\n            calc(" + verticalSpacing + "px " + ANCHOR_OFFSET[verticalAnchor] + ")\n          )"
    }, style)
  }, children)));
}

var _excluded$a = ["wrapperStyle"];
function Svg(_ref) {
  var _ref$wrapperStyle = _ref.wrapperStyle,
      wrapperStyle = _ref$wrapperStyle === void 0 ? {} : _ref$wrapperStyle,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$a);

  useForceRender();
  var ref = useRef(null);
  var tooltipParent = ref.current;
  return React.createElement("div", {
    ref: ref,
    style: _extends({}, wrapperStyle, {
      position: 'relative'
    })
  }, tooltipParent && React.createElement(TooltipParentProvider, {
    value: tooltipParent
  }, React.createElement("svg", Object.assign({}, props))));
}

var _excluded$b = ["x-data", "y-data"];
function Texts(_ref) {
  var dataX = _ref['x-data'],
      dataY = _ref['y-data'],
      props = _objectWithoutPropertiesLoose(_ref, _excluded$b);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var x = createDataScale(dataX, xScale, 0, 'center');
  var y = createDataScale(dataY, yScale, 0, 'center');
  return React.createElement(Elements, Object.assign({}, props, {
    tag: "text",
    x: combineGetters([x, props.x]),
    y: combineGetters([y, props.y])
  }));
}

var _excluded$c = ["data", "x", "y", "curve", "dataKey", "enter", "duration", "delay", "easing"];

function sanitizeYConfig(config, scale) {
  if (typeof config === 'function') {
    return {
      to: config,
      base: function base() {
        return scaleZero(scale);
      }
    };
  }

  var base = config.base,
      to = config.to;
  return {
    to: to,
    base: toAccessor(base)
  };
}
/** @deprecated since v0.2.0. Use `Area` instead. Will be removed in v1.0.0 */


function AreaData(_ref) {
  var data = _ref.data,
      x = _ref.x,
      _y = _ref.y,
      _ref$curve = _ref.curve,
      curve = _ref$curve === void 0 ? 'line' : _ref$curve,
      dataKey = _ref.dataKey,
      enter = _ref.enter,
      duration = _ref.duration,
      delay = _ref.delay,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$c);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var style = useCascadingStyle(props);
  var curveFn = buildCurveFn(curve);
  var animation = useSanitizedCascadingAnimation({
    duration: duration,
    delay: delay,
    easing: easing
  });
  var y = sanitizeYConfig(_y, yScale);

  var d = function d(data) {
    var _area$x$y0$y1$curve;

    var points = data.map(function (d) {
      return [computePos(x(d), xScale), computePos(y.base(d), yScale), computePos(y.to(d), yScale)];
    }).filter(function (d) {
      return d.every(function (p) {
        return !isNaN(p);
      });
    });
    return (_area$x$y0$y1$curve = area().x(function (d) {
      return d[0];
    }).y0(function (d) {
      return d[1];
    }).y1(function (d) {
      return d[2];
    }).curve(curveFn)(points)) != null ? _area$x$y0$y1$curve : '';
  };

  return React.createElement(AnimatedDataset, Object.assign({
    tag: "path",
    dataset: [data],
    attrs: _extends({
      d: d,
      fill: 'none'
    }, style),
    keyFn: dataKey,
    init: enter
  }, animation));
}

var _excluded$d = ["fn", "resolution", "fnBase"];
/** @deprecated since v0.2.0. Will be removed in v1.0.0 */

function AreaFunction(_ref) {
  var fn = _ref.fn,
      _ref$resolution = _ref.resolution,
      resolution = _ref$resolution === void 0 ? 1 : _ref$resolution,
      _ref$fnBase = _ref.fnBase,
      fnBase = _ref$fnBase === void 0 ? function () {
    return 0;
  } : _ref$fnBase,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$d);

  var _useChartContext = useChartContext(),
      left = _useChartContext.left,
      right = _useChartContext.right;

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale;

  if (!isScaleContinuous(xScale)) throw new Error('AreaFunction must be used in a Cartesian component with a continuous x scale');
  var step = 1 / resolution;
  var xValues = range(left, right + step, step).map(xScale.invert);
  return React.createElement(AreaData, Object.assign({
    data: xValues,
    x: function x(_x) {
      return _x;
    },
    y: {
      to: fn,
      base: fnBase
    }
  }, props));
}

function splitStyleProps(props) {
  var keys = Object.keys(props);
  var existingKeys = keys.filter(function (k) {
    return typeof k !== 'undefined';
  });

  var _partition = partition(existingKeys, function (key) {
    return typeof props[key] === 'function';
  }),
      fnKeys = _partition[0],
      valKeys = _partition[1];

  return tuple(pick(props, valKeys), pick(props, fnKeys));
}
/** @deprecated */


function useComputableStyle(curr, colorScale) {
  var _splitStyleProps = splitStyleProps(curr),
      props = _splitStyleProps[0],
      getters = _splitStyleProps[1];

  var style = useCascadingStyle(props);
  var getterKeys = Object.keys(getters);

  if (getterKeys.length > 0 && !colorScale) {
    throw new Error("Define a color scale");
  }

  return _extends({}, style, mapValues(getters, function (fn) {
    return function (d) {
      return colorScale(fn(d));
    };
  }));
}

var _excluded$e = ["data", "x", "y", "dataKey", "delay", "duration", "enter", "easing"];

function sanitizeAxisConfig$1(config, scale) {
  if (typeof config === 'function') {
    return {
      to: config,
      base: config
    };
  }

  var to = config.to,
      _config$base = config.base,
      base = _config$base === void 0 ? function () {
    return scaleZero(scale);
  } : _config$base;
  return {
    to: toAccessor(to),
    base: toAccessor(base)
  };
}
/** @deprecated since v0.2.0. Use `Bars` instead. Will be removed in v1.0.0 */


function BarData(_ref) {
  var data = _ref.data,
      _x = _ref.x,
      _y = _ref.y,
      dataKey = _ref.dataKey,
      delay = _ref.delay,
      duration = _ref.duration,
      _ref$enter = _ref.enter,
      enter = _ref$enter === void 0 ? {} : _ref$enter,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$e);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale,
      colorScale = _useCartesianContext.colorScale;

  var style = useComputableStyle(props, colorScale);
  var x = sanitizeAxisConfig$1(_x, xScale);
  var y = sanitizeAxisConfig$1(_y, yScale);
  var animation = useSanitizedCascadingAnimation({
    delay: delay,
    duration: duration,
    easing: easing
  });

  var x1 = function x1(d) {
    return computePos(x.base(d), xScale, 'start');
  };

  var y1 = function y1(d) {
    return computePos(y.base(d), yScale, 'end');
  };

  var x2 = function x2(d) {
    return computePos(x.to(d), xScale, 'end');
  };

  var y2 = function y2(d) {
    return computePos(y.to(d), yScale, 'start');
  };

  return React.createElement(AnimatedDataset, Object.assign({
    tag: "rect",
    dataset: data,
    attrs: _extends({
      x: function x(d) {
        return Math.min(x1(d), x2(d));
      },
      y: function y(d) {
        return Math.min(y1(d), y2(d));
      },
      width: function width(d) {
        return Math.abs(x1(d) - x2(d));
      },
      height: function height(d) {
        return Math.abs(y1(d) - y2(d));
      }
    }, style),
    init: enter,
    keyFn: function keyFn(d, i) {
      var _dataKey;

      return (_dataKey = dataKey == null ? void 0 : dataKey(d, i)) != null ? _dataKey : i;
    }
  }, animation));
}

/** @deprecated */

function buildComputePos(dataAccessor, positionGetter, offsetGetter, scale, debugName) {
  return function (datum, position) {
    if (position === void 0) {
      position = 'center';
    }

    var offset = offsetGetter !== undefined ? toAccessor(offsetGetter)(datum) : 0;

    if (positionGetter !== undefined) {
      var positionAccessor = toAccessor(positionGetter);
      return positionAccessor(datum) + offset;
    }

    if (dataAccessor !== undefined) {
      var value = dataAccessor(datum);
      return computePos(value, scale, position) + offset;
    }

    throw new Error("Please declare 1 of data" + debugName + " or position" + debugName);
  };
}

var _excluded$f = ["data", "dataX", "dataY", "dataKey", "positionX", "positionY", "offsetX", "offsetY", "text", "enter", "delay", "duration"];
/** @deprecated since v0.2.0. Use `Texts` instead. Will be removed in v1.0.0 */

function LabelsData(_ref) {
  var data = _ref.data,
      dataX = _ref.dataX,
      dataY = _ref.dataY,
      dataKey = _ref.dataKey,
      positionX = _ref.positionX,
      positionY = _ref.positionY,
      offsetX = _ref.offsetX,
      offsetY = _ref.offsetY,
      text = _ref.text,
      _ref$enter = _ref.enter,
      enter = _ref$enter === void 0 ? {} : _ref$enter,
      delay = _ref.delay,
      duration = _ref.duration,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$f);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale,
      colorScale = _useCartesianContext.colorScale;

  var style = useComputableStyle(props, colorScale);
  var animation = useSanitizedCascadingAnimation({
    delay: delay,
    duration: duration
  });
  var computeXPos = buildComputePos(dataX, positionX, offsetX, xScale, 'X');
  var computeYPos = buildComputePos(dataY, positionY, offsetY, yScale, 'Y');
  return React.createElement(AnimatedDataset, Object.assign({
    tag: "text",
    dataset: data,
    attrs: _extends({
      x: function x(d) {
        return computeXPos(d, 'center');
      },
      y: function y(d) {
        return computeYPos(d, 'center');
      },
      text: text
    }, style),
    init: enter,
    keyFn: function keyFn(d, i) {
      var _dataKey;

      return (_dataKey = dataKey == null ? void 0 : dataKey(d, i)) != null ? _dataKey : i;
    }
  }, animation));
}

var _excluded$g = ["data", "x", "y", "dataKey", "curve", "delay", "duration", "enter", "easing"];
/** @deprecated since v0.2.0. Use `Line` instead. Will be removed in v1.0.0 */

function LineData(_ref) {
  var data = _ref.data,
      x = _ref.x,
      y = _ref.y,
      dataKey = _ref.dataKey,
      _ref$curve = _ref.curve,
      type = _ref$curve === void 0 ? 'line' : _ref$curve,
      delay = _ref.delay,
      duration = _ref.duration,
      enter = _ref.enter,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$g);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale;

  var style = useCascadingStyle(props);
  var curveFn = buildCurveFn(type);
  var animation = useSanitizedCascadingAnimation({
    duration: duration,
    delay: delay,
    easing: easing
  });

  var d = function d(data) {
    var _line$curve;

    var points = data.map(function (d) {
      return tuple(computePos(x(d), xScale), computePos(y(d), yScale));
    }).filter(function (d) {
      return d.every(function (p) {
        return !isNaN(p);
      });
    });
    return (_line$curve = line$1().curve(curveFn)(points)) != null ? _line$curve : '';
  };

  return React.createElement(AnimatedDataset, Object.assign({
    tag: "path",
    dataset: [data],
    attrs: _extends({
      d: d,
      fill: 'none'
    }, style),
    init: enter,
    keyFn: dataKey
  }, animation));
}

var _excluded$h = ["fn", "resolution"];
/** @deprecated since v0.2.0. Will be removed in v1.0.0 */

function LineFunction(_ref) {
  var fn = _ref.fn,
      _ref$resolution = _ref.resolution,
      resolution = _ref$resolution === void 0 ? 1 : _ref$resolution,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$h);

  var _useChartContext = useChartContext(),
      left = _useChartContext.left,
      right = _useChartContext.right;

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale;

  if (!isScaleContinuous(xScale)) throw new Error('LineFunction must be used in a Cartesian component with a continuous x scale');
  var step = 1 / resolution;
  var xValues = range(left, right + step, step).map(xScale.invert);
  return React.createElement(LineData, Object.assign({
    data: xValues,
    x: function x(_x) {
      return _x;
    },
    y: fn
  }, props));
}

var _excluded$i = ["data", "x", "y", "r", "dataKey", "duration", "delay", "enter", "easing"];
/** @deprecated since v0.2.0. Use `Circles` instead. Will be removed in v1.0.0 */

function PointData(_ref) {
  var data = _ref.data,
      x = _ref.x,
      y = _ref.y,
      _ref$r = _ref.r,
      _r = _ref$r === void 0 ? 1 : _ref$r,
      dataKey = _ref.dataKey,
      duration = _ref.duration,
      delay = _ref.delay,
      enter = _ref.enter,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$i);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale,
      colorScale = _useCartesianContext.colorScale;

  var style = useComputableStyle(props, colorScale);
  var animation = useSanitizedCascadingAnimation({
    delay: delay,
    duration: duration,
    easing: easing
  });
  return React.createElement(AnimatedDataset, Object.assign({
    tag: "circle",
    dataset: data,
    keyFn: function keyFn(d, i) {
      var _dataKey;

      return (_dataKey = dataKey == null ? void 0 : dataKey(d, i)) != null ? _dataKey : i;
    },
    attrs: _extends({
      cx: function cx(d) {
        return computePos(x(d), xScale);
      },
      cy: function cy(d) {
        return computePos(y(d), yScale);
      },
      r: function r(d) {
        return typeof _r === 'function' ? _r(d) : _r;
      }
    }, style),
    init: enter
  }, animation));
}

var _excluded$j = ["data", "x", "y", "width", "height", "dataKey", "duration", "delay", "enter", "easing"];
/** @deprecated since v0.2.0. Use `Rects` instead. Will be removed in v1.0.0 */

function RectData(_ref) {
  var data = _ref.data,
      _x = _ref.x,
      _y = _ref.y,
      _width = _ref.width,
      _height = _ref.height,
      dataKey = _ref.dataKey,
      duration = _ref.duration,
      delay = _ref.delay,
      enter = _ref.enter,
      easing = _ref.easing,
      props = _objectWithoutPropertiesLoose(_ref, _excluded$j);

  var _useCartesianContext = useCartesianContext(),
      xScale = _useCartesianContext.xScale,
      yScale = _useCartesianContext.yScale,
      colorScale = _useCartesianContext.colorScale;

  var style = useComputableStyle(props, colorScale);
  var animation = useSanitizedCascadingAnimation({
    delay: delay,
    duration: duration,
    easing: easing
  });

  var _x2 = function x(d) {
    return computePos(_x(d), xScale);
  };

  var _y2 = function y(d) {
    return computePos(_y(d), yScale);
  }; // TODO: refactor. _width can be either a px value
  // or a function that returns a data value


  var width = function width(d) {
    return typeof _width === 'function' ? Math.abs(computePos(toAccessor(_width)(d), xScale)) - computePos(scaleZero(xScale), xScale) : _width;
  };

  var height = function height(d) {
    return typeof _height === 'function' ? Math.abs(computePos(toAccessor(_height)(d), yScale) - computePos(scaleZero(yScale), yScale)) : _height;
  };

  return React.createElement(AnimatedDataset, Object.assign({
    tag: "rect",
    dataset: data,
    keyFn: function keyFn(d, i) {
      var _dataKey;

      return (_dataKey = dataKey == null ? void 0 : dataKey(d, i)) != null ? _dataKey : i;
    },
    attrs: _extends({
      x: function x(d) {
        return _x2(d) - width(d) / 2;
      },
      y: function y(d) {
        return _y2(d) - height(d) / 2;
      },
      width: width,
      height: height
    }, style),
    init: enter
  }, animation));
}

var wideToNarrow = function wideToNarrow(data, cols, namesTo, valuesTo) {
  return data.flatMap(function (datum) {
    var datumWithoutCols = omit(datum, cols);
    return cols.map(function (c) {
      var _extends2;

      return _extends({}, datumWithoutCols, (_extends2 = {}, _extends2[namesTo] = c, _extends2[valuesTo] = datum[c], _extends2));
    });
  });
}; // export const narrowToWide = <T extends object, C extends keyof T, V extends keyof T>(
//   data: T[],
//   key: (d: T) => string,
//   col: C,
//   value: V
// ) => {
//   const groupedBy = groupBy(data, key)
//   return Object.values(groupedBy).map((arr) => {
//     const datumWithoutColAndValue = omit(arr[0], [col, value])
//     return arr.reduce((acc, curr) => {
//       acc[curr[col as string]] = curr[value]
//       return acc
//     }, datumWithoutColAndValue)
//   })
// }

var stackNarrow = function stackNarrow(_ref) {
  var data = _ref.data,
      categories = _ref.categories,
      getCategory = _ref.getCategory,
      getGroup = _ref.getGroup,
      getValue = _ref.getValue;
  var groups = Object.values(groupBy(data, getGroup));
  return groups.flatMap(function (group) {
    var dataByCategory = keyBy(group, getCategory);
    var dataSortedByCategory = compact(categories.map(function (category) {
      return get(dataByCategory, category, undefined);
    }));
    var top = 0;
    var bottom = 0;
    return dataSortedByCategory.map(function (datum) {
      var value = getValue(datum);
      var base = value >= 0 ? top : bottom;
      var to = base + value;

      if (value >= 0) {
        top = to;
      } else {
        bottom = to;
      }

      return {
        datum: datum,
        base: base,
        to: to,
        value: value,
        group: getGroup(datum),
        category: getCategory(datum)
      };
    });
  });
};

export { AnimatedDataset, Animation, AnimationConsumer, AnimationContext, Area, AreaData, AreaFunction, BarData, Bars, Cartesian, CartesianConsumer, CartesianContext, Chart, ChartConsumer, ChartContext, Circles, ClipChart, ClipRect, Elements, Grid, GridConsumer, GridContext, LabelsData, Line, LineData, LineFunction, MouseConsumer, MouseContext, PointData, Portal, Range, RectData, Rects, Style, StyleConsumer, StyleContext, Svg, Texts, Tooltip, TooltipParentProvider, buildComputePos, combineGetters, computePos, createDataScale, element, isAccessor, isNotNil, isScaleCategorical, isScaleContinuous, isUndefinedTuple, scaleStep, scaleZero, stackNarrow, strTuple, toAccessor, tuple, useAnimationContext, useCartesianContext, useCascadingAnimation, useCascadingStyle, useChartContext, useComputableStyle, useGridContext, useMouseContext, useSanitizedCascadingAnimation, useStyleContext, useTooltipParent, wideToNarrow, wrapIteratee };
//# sourceMappingURL=react-composable-charts.esm.js.map
